module Emitter where

import List

import Parser
import Table
import Instruction

data Environment = Environment (Table (Type, Int), Int, Int) deriving Show

envTable (Environment (table, offset, label)) = table
envOffset (Environment (table, offset, label)) = offset
envLabel (Environment (table, offset, label)) = label
incEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset - 1, label))
decEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset + 1, label))
incEnvLabel (Environment (table, offset, label)) = (Environment (table, offset , label + 1))

data EmitValue =
    EmitVoidValue |
    EmitIntValueTemporary Int |
    EmitIntVariable Int String |
    EmitIntValueConstant Int |
    EmitValueFalseLabels [String] |
    EmitValueTrueLabels [String] |
    EmitIntValueRegisterA |
    EmitIntValueRegisterY |
    EmitAddress Int String |
    EmitAddressRegisterA |
    EmitAddressRegisterY |
    EmitNone deriving (Show)

data EmitResult =
    EmitResult (EmitValue, [(Instruction, String)], Environment) |
    EmitResultAlternative EmitResult [EmitResult]
    deriving Show

emitResultValue (EmitResult (ev, code, env)) = ev
emitResultValue (EmitResultAlternative er _) = emitResultValue er
emitResultCode (EmitResult (ev, code, env)) = code
emitResultCode (EmitResultAlternative er _) = emitResultCode er
emitResultEnvironment (EmitResult (ev, code, env)) = env
emitResultEnvironment (EmitResultAlternative er _) = emitResultEnvironment er

emitResultSelect p er@(EmitResult _) = er
emitResultSelect p (EmitResultAlternative der (er:ers)) | p er = er
emitResultSelect p (EmitResultAlternative der (er:ers)) = emitResultSelect p (EmitResultAlternative der ers)
emitResultSelect p (EmitResultAlternative der []) = der

emitResultIsFalseLabels (EmitResult (EmitValueFalseLabels _, _, _)) = True
emitResultIsFalseLabels _ = False

emitResultIsTrueLabels (EmitResult (EmitValueTrueLabels _, _, _)) = True
emitResultIsTrueLabels _ = False

data EmitContext = EmitContext { emitContextBreakLabel::String, emitContextNeedValue::Bool }

emitError = error . ("emitter error: " ++)

emit tree = emitProgram tree
            (Environment (initialTable "undeclared variable ", 
                          if isCalledSetLedBinaryProgram tree then 0xd else 0xf, 0))

emitProgram (Program functions) env = emitConcat emitFunction env functions
emitConcat f env l = concat $ map (\x -> f x env) l

emitFunction (Function (name, ftype, arguments, block)) env =
    emitResultCode (emitBlock block env (EmitContext {emitContextBreakLabel = "", emitContextNeedValue = False}))

emitBlock (Block (declarations, statements)) env context =
    EmitResult(emitResultValue statementsResult,
               emitResultCode statementsResult,
               Environment (envTable env, envOffset env, envLabel (emitResultEnvironment statementsResult)))
    where
      newEnv = emitDeclarations declarations env
      statementsResult = emitStatements statements newEnv context

emitStatements = emitSequence emitStatement

emitSequence ef (t:ts) env context = 
    EmitResult (if ts == [] then EmitNone else emitResultValue emitResult,
                (emitResultCode emitResult) ++ emitResultCode nextEmitResult,
                emitResultEnvironment nextEmitResult)
    where
      emitResult = ef t env context
      nextEmitResult = emitSequence ef ts (emitResultEnvironment emitResult) context
emitSequence ef [] env context = EmitResult (EmitNone, [], env)

emitStatement tree env context = 
    case tree of
      (If (expression, statements)) -> emitStatementIf (emitExpression expression env (context {emitContextNeedValue = True}))
                                       statements context

      (IfElse (expression, thenStatements, elseStatements)) -> emitStatementIfElse (emitExpression expression env (context {emitContextNeedValue = True}))
                                                               thenStatements elseStatements context

      (While (expression, statements)) -> emitStatementWhile (emitExpression expression env (context {emitContextNeedValue = True}))
                                          statements context

      (InfiniteLoop (statements)) ->
          let
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
              emitResult = emitStatement statements env
                           (context {emitContextBreakLabel = label2})
              newEnv = emitResultEnvironment emitResult
          in
            EmitResult(EmitNone,
                       [(Label label1, "infinite loop start")] ++
                       emitResultCode emitResult ++
                       [(FlagResetDontCare, ""),
                        (OperandString1 "JUMP" label1, "infinite loop jump"),
                        (Label label2, "")],
                       incEnvLabel newEnv)

      (Break) -> 
          EmitResult(EmitNone,
                     [(FlagResetDontCare, ""),
                      (OperandString1 "JUMP" (emitContextBreakLabel context), "break")],
                     env)

      (StatementExpression expression) -> emitExpression expression env context

      (StatementBlock block) -> emitBlock block env context

      (StatementEmpty) -> EmitResult (EmitNone, [], env)

      _ -> emitError ("unkown statement")

emitStatementIf (EmitResult (condVal, condCode, condEnv)) statements context =
    let
        thenResult = emitStatement statements condEnv context
        newEnv = emitResultEnvironment thenResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitNone,
       condCode ++ emitValueToRegisterA condVal ++
       [(OperandInt1 "AIA" 0xf, "if judege"),
        (OperandInt1 "AIA" 1, ""),
        (OperandString1 "JUMP" label1, "")] ++ 
       emitResultCode thenResult ++
       [(Label label1, "if end")],
       incEnvLabel newEnv)
emitStatementIf (EmitResultAlternative der (EmitResult (EmitValueFalseLabels condLabels, condCode, condEnv):ers)) statements context =
    let
        thenResult = emitStatement statements condEnv context
        newEnv = emitResultEnvironment thenResult
    in                      
      EmitResult
      (EmitNone,
       condCode ++
       emitResultCode thenResult ++
       emitLabels condLabels "if end",
       newEnv)
emitStatementIf (EmitResultAlternative der []) statements context = emitStatementIf der statements context
emitStatementIf (EmitResultAlternative der (er:ers)) statements context = emitStatementIf (EmitResultAlternative der ers) statements context

emitStatementIfElse (EmitResult (condVal, condCode, condEnv)) thenStatements elseStatements context =
    let
        thenResult = emitStatement thenStatements condEnv context
        elseResult = emitStatement elseStatements (emitResultEnvironment thenResult) context
        newEnv = emitResultEnvironment elseResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      EmitResult
      (EmitNone,
       condCode ++ emitValueToRegisterA condVal ++
       [(OperandInt1 "CIA" 0, "if else judge"),
        (OperandString1 "JUMP" label1, "")] ++ 
       emitResultCode elseResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label2, "if else jump"),
        (Label label1, "")] ++
       emitResultCode thenResult ++
       [(Label label2, "if else end")],
       (incEnvLabel . incEnvLabel) newEnv)
emitStatementIfElse (EmitResultAlternative der (EmitResult (EmitValueFalseLabels condLabels, condCode, condEnv):ers)) thenStatements elseStatements context =
    let
        thenResult = emitStatement thenStatements condEnv context
        elseResult = emitStatement elseStatements (emitResultEnvironment thenResult) context
        newEnv = emitResultEnvironment elseResult
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      EmitResult
      (EmitNone,
       condCode ++
       emitResultCode thenResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "if else jump")] ++
       emitLabels condLabels "else" ++ 
       emitResultCode elseResult ++
       [(Label label1, "if else end")],
       incEnvLabel newEnv)
emitStatementIfElse (EmitResultAlternative der []) thenStatements elseStatements context = emitStatementIfElse der thenStatements elseStatements context
emitStatementIfElse (EmitResultAlternative der (er:ers)) thenStatements elseStatements context = emitStatementIfElse (EmitResultAlternative der ers) thenStatements elseStatements context

emitStatementWhile (EmitResult (condVal, condCode, condEnv)) statements context =
    let
        labelNo = envLabel condEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
        contentResult = emitStatement statements ((incEnvLabel . incEnvLabel) condEnv)
                        (context {emitContextBreakLabel = label2})
        newEnv = emitResultEnvironment contentResult
    in
      EmitResult
      (EmitNone,
       [(Label label1, "while start")] ++
       condCode ++ emitValueToRegisterA condVal ++
       [(OperandInt1 "AIA" 0xf, "while judege"),
        (OperandInt1 "AIA" 1, ""),
        (OperandString1 "JUMP" label2, "")] ++
       emitResultCode contentResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "while loop jump"),
        (Label label2, "while end")],
       newEnv)
emitStatementWhile (EmitResultAlternative der (EmitResult (EmitValueFalseLabels condLabels, condCode, condEnv):ers)) statements context =
    let
        labelNo = envLabel condEnv
        label1 = emitLabel labelNo
        contentResult = emitStatement statements (incEnvLabel condEnv)
                        (context {emitContextBreakLabel = (head condLabels)})
        newEnv = emitResultEnvironment contentResult
    in
      EmitResult
      (EmitNone,
       [(Label label1, "while start")] ++
       condCode ++ 
       emitResultCode contentResult ++
       [(FlagResetDontCare, ""),
        (OperandString1 "JUMP" label1, "while loop jump")] ++
       emitLabels condLabels "while end",
       incEnvLabel newEnv)
emitStatementWhile (EmitResultAlternative der []) statements context = emitStatementWhile der statements context
emitStatementWhile (EmitResultAlternative der (er:ers)) statements context = emitStatementWhile (EmitResultAlternative der ers) statements context

emitExpression tree env context = 
    case tree of
      (Number i) -> EmitResult (EmitIntValueConstant i, [], env)
      (VariableValue s) ->
          case contentTable (envTable env) s of
            (TypeInt, offset) -> EmitResult (EmitIntVariable offset s, [], env)
            (TypePointer, address) -> EmitResult (EmitAddress address s, [], env)

      PreOperation1 ("+" , expression) -> emitExpression expression env context

      PreOperation1 ("-" , (Number i)) -> EmitResult (EmitIntValueConstant (immediateValue (-i)), [], env)

      PreOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitLValue expression env context
                  newEnv = emitResultEnvironment emitResult
              in
                EmitResult(EmitIntValueRegisterA,
                           emitResultCode emitResult ++ 
                           emitAddressToRegisterY (emitResultValue emitResult) ++
                           [(Operand0 "MA", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Operator '" ++ operator ++ "'"),
                            (Operand0 "AM", "")],
                           newEnv)

      PreOperation1 ("&" , expression) ->
          let
              emitResult = emitLValue expression env context
              newEnv = emitResultEnvironment emitResult
          in
            EmitResult(EmitIntValueRegisterY,
                       emitResultCode emitResult ++
                       (emitAddressToRegisterY (emitResultValue emitResult)),
                       newEnv)
                                          
      PreOperation1 ("!" , expression) -> emitExpressionNot (emitExpression expression env context) context

      PreOperation1 (operator , expression) ->
          let
              emitResult = emitExpression expression env context
              newEnv = emitResultEnvironment emitResult
          in
            case operator of
              "-" ->
                  EmitResult(EmitIntValueRegisterA,
                             emitResultCode emitResult ++
                             (emitValueToRegisterA (emitResultValue emitResult)) ++
                             [(FlagResetRegistersSafe, ""),
                              (OperandString1 "CAL" "CMPL", "Unary operator '" ++ operator ++ "'"),
                              (OperandInt1 "AIA" 1, "")],
                             newEnv)
              _ -> emitError ("Operator " ++ operator ++ " not supported")
                  
      PostOperation1 (operator , expression)
          | any (==operator) ["++", "--"] ->
              let
                  emitResult = emitLValue expression env context
                  newEnv = emitResultEnvironment emitResult
              in
                EmitResult(if emitContextNeedValue context then EmitIntValueRegisterA else EmitNone,
                           emitResultCode emitResult ++
                           emitAddressToRegisterY (emitResultValue emitResult) ++
                           [(Operand0 "MA", ""),
                            (OperandInt1 "AIA" (if operator == "++" then 1 else 0xf), "Operator '" ++ operator ++ "'"),
                            (Operand0 "AM", "")] ++
                           (if emitContextNeedValue context
                            then [(OperandInt1 "AIA" (if operator == "++" then 0xf else 1), "cancel")]
                            else []
                           ),
                           newEnv)
      PostOperation1 (operator , expression) -> emitError ("unary operator " ++ operator ++ " not supported")

      Operation2 ("&&" , expression1, expression2) -> emitExpressionAnd expression1 expression2 env context
      Operation2 ("||" , expression1, expression2) -> emitExpressionOr expression1 expression2 env context
                                                

      Operation2 (operator , expression1, expression2) | any (operator==) ["=", "+=", "-="] ->
          let
              emitResult2 = emitExpression expression2 env (context {emitContextNeedValue = True})
              emitResult1 = emitLValue expression1 (incEnvOffset (emitResultEnvironment emitResult2)) (context {emitContextNeedValue = True})
              newEnv = decEnvOffset (emitResultEnvironment emitResult1)
              offset = envOffset newEnv
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            case emitResultCode emitResult1 of
              [] ->
                  EmitResult(EmitIntValueRegisterA,
                             emitResultCode emitResult2 ++
                             emitValueToRegisterA (emitResultValue emitResult2) ++
                             emitAddressToRegisterY (emitResultValue emitResult1) ++
                             (if operator == "+=" || operator == "-="
                              then
                                  [(Operand0 ("M" ++ [head operator]), "Operator '" ++ operator ++ "'"),
                                   (Operand0 "AM", "")]
                              else
                                  [(Operand0 "AM", "Operator '" ++ operator ++ "'")]),
                             newEnv)
              _ ->
                  EmitResult(EmitIntValueRegisterA,
                             emitResultCode emitResult2 ++
                             emitValueToRegisterA (emitResultValue emitResult2) ++
                             [(OperandInt1 "TIY" offset, ""),
                              (Operand0 "AM", "")] ++
                             emitResultCode emitResult1 ++ 
                             emitAddressToRegisterY (emitResultValue emitResult1) ++
                             [(Operand0 "CH", ""),
                              (OperandInt1 "TIY" offset, ""),
                              (Operand0 "MA", ""),
                              (OperandInt1 "TIY" 0xf, ""),
                              (Operand0 "CH", "copy Br to Yr (too long)"),
                              (Label label2, ""),
                              (OperandInt1 "AIY" 1, ""),
                              (OperandString1 "JUMP" label1, ""),
                              (Operand0 "CH", ""),
                              (OperandInt1 "AIY" 0xf, ""),
                              (Operand0 "CH", ""),
                              (OperandString1 "JUMP" label2, ""),
                              (Label label1, ""),
                              (Operand0 "CH", "")] ++
                              (if operator == "+=" || operator == "-="
                               then
                                   [(Operand0 ("M" ++ [head operator]), "Operator '" ++ operator ++ "'"),
                                    (Operand0 "AM", "")]
                               else
                                   [(Operand0 "AM", "Operator '" ++ operator ++ "'")]),
                              (incEnvLabel.incEnvLabel) newEnv)

      Operation2 (operator , expression1, expression2) ->
          let
              emitResult1 = emitExpression expression1 env context
              emitResult2 = emitExpression expression2 (incEnvOffset (emitResultEnvironment emitResult1)) context
              newEnv = decEnvOffset (emitResultEnvironment emitResult2)
              offset = envOffset newEnv

              reverseEmitResult1 = emitExpression expression2 env context
              reverseEmitResult2 = emitExpression expression1 (incEnvOffset (emitResultEnvironment reverseEmitResult1)) context
              reverseNewEnv = decEnvOffset (emitResultEnvironment reverseEmitResult2)
              reverseOffset = envOffset reverseNewEnv
          in
            case operator of
              o | any (o==) ["+", "-", "!="] -> let op = if o == "!=" then "-" else o in
                  EmitResult(EmitIntValueRegisterA,
                             emitResultCode emitResult1 ++
                             (emitValueToRegisterA (emitResultValue emitResult1)) ++
                             [(OperandInt1 "TIY" offset, ""),
                              (Operand0 "AM", "")] ++
                             emitResultCode emitResult2 ++
                             (emitValueToRegisterA (emitResultValue emitResult2)) ++
                             [(OperandInt1 "TIY" offset, ""),
                              (Operand0 ("M" ++ op), "Operator '" ++ operator ++ "'")],
                             newEnv)

              "," ->
                  EmitResult(emitResultValue emitResult2,
                             emitResultCode emitResult1 ++
                             emitResultCode emitResult2,
                             newEnv)

              o | any (o==) [">=", "<=", ">", "<"] ->
                  let
                      emitResultOp1 = (if o == ">=" || o == "<" then emitResult1 else reverseEmitResult1)
                      emitResultOp2 = (if o == ">=" || o == "<" then emitResult2 else reverseEmitResult2)
                      targetNewEnv = (if o == ">=" || o == "<" then newEnv else reverseNewEnv)
                                      
                      labelNo = envLabel targetNewEnv
                      label1 = emitLabel labelNo
                      label2 = emitLabel (labelNo + 1)
                  in
                    EmitResultAlternative
                    (
                     EmitResult
                     (
                      EmitIntValueRegisterA,
                      emitResultCode emitResultOp1 ++
                      (emitValueToRegisterA (emitResultValue emitResultOp1)) ++
                      [(OperandInt1 "TIY" offset, ""),
                       (Operand0 "AM", "")] ++
                      emitResultCode emitResultOp2 ++
                      (emitValueToRegisterA (emitResultValue emitResultOp2)) ++
                      [(OperandInt1 "TIY" offset, ""),
                       (Operand0 "CH", ""),
                       (if o == "<=" || o == ">="
                        then (OperandInt1 "TIA" 0, "set false")
                        else (OperandInt1 "TIA" 1, "set true")
                       ),
                       (Operand0 "CH", ""),
                       (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                       (OperandString1 "JUMP" label1, ""),
                       (if o == "<=" || o == ">="
                        then (OperandInt1 "TIA" 1, "set true")
                        else (OperandInt1 "TIA" 0, "set false")
                       ),
                       (Operand0 "CH", ""),
                       (Label label1, ""),
                       (Operand0 "CH", "")],
                      incEnvLabel targetNewEnv
                     )
                    )

                    (emitResultCodeSizeSort
                     [
                      EmitResult
                      (
                       EmitValueFalseLabels [label1],
                       emitResultCode emitResultOp1 ++
                       (emitValueToRegisterA (emitResultValue emitResultOp1)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 "AM", "")] ++
                       emitResultCode emitResultOp2 ++
                       (emitValueToRegisterA (emitResultValue emitResultOp2)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                       (if o == "<" || o == ">" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                       [(OperandString1 "JUMP" label1, "OOCV")] ++
                       (if o == "<" || o == ">" then [(Label label2, "")]  else []),
                       (incEnvLabel.incEnvLabel) targetNewEnv
                      ),
                      EmitResult
                      (
                       EmitValueTrueLabels [label1],
                       emitResultCode emitResultOp1 ++
                       (emitValueToRegisterA (emitResultValue emitResultOp1)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 "AM", "")] ++
                       emitResultCode emitResultOp2 ++
                       (emitValueToRegisterA (emitResultValue emitResultOp2)) ++
                       [(OperandInt1 "TIY" offset, ""),
                        (Operand0 ("M-"), "Operator '" ++ operator ++ "'")] ++
                       (if o == "<=" || o == ">=" then [(OperandString1 "JUMP" label2, "")]  else []) ++
                       [(OperandString1 "JUMP" label1, "OOCV")] ++
                       (if o == "<=" || o == ">=" then [(Label label2, "")]  else []),
                       (incEnvLabel.incEnvLabel) targetNewEnv
                      )
                     ])
              "==" ->
                  EmitResultAlternative
                  (
                   EmitResult
                   (
                    EmitIntValueRegisterA,
                    emitResultCode emitResult1 ++
                    (emitValueToRegisterA (emitResultValue emitResult1)) ++
                    [(OperandInt1 "TIY" offset, ""),
                     (Operand0 "AM", "")] ++
                    emitResultCode emitResult2 ++
                    (emitValueToRegisterA (emitResultValue emitResult2)) ++
                    [(OperandInt1 "TIY" offset, ""),
                     (FlagResetRegistersSafe, ""),
                     (OperandString1 "CAL" "CHNG", "set false Ar'"),
                     (OperandInt1 "TIA" 0, ""),
                     (OperandInt1 "TIY" offset, ""),
                     (FlagResetRegistersSafe, ""),
                     (OperandString1 "CAL" "CHNG", ""),
                     (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                     (OperandInt1 "AIA" 0xf, ""),
                     (OperandString1 "CAL" "CHNG", "")],
                    newEnv
                   )
                  )
                  [
                   EmitResult
                   (
                    EmitValueFalseLabels [(emitLabel (envLabel newEnv))],
                    case (emitResultValue emitResult1, emitResultValue emitResult2) of
                      ((EmitIntValueConstant constant), _) ->
                          emitResultCode emitResult2 ++
                          (emitValueToRegisterA (emitResultValue emitResult2)) ++
                          [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                           (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "")]
                      (_, (EmitIntValueConstant constant)) ->
                          emitResultCode emitResult1 ++
                          (emitValueToRegisterA (emitResultValue emitResult1)) ++
                          [(OperandInt1 "CIA" constant, "Compare constant " ++ show constant ++ " OOCMP=="),
                           (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "")]
                      _ ->
                          emitResultCode emitResult1 ++
                          (emitValueToRegisterA (emitResultValue emitResult1)) ++
                          [(OperandInt1 "TIY" offset, ""),
                           (Operand0 "AM", "")] ++
                          emitResultCode emitResult2 ++
                          (emitValueToRegisterA (emitResultValue emitResult2)) ++
                          [(OperandInt1 "TIY" offset, ""),
                           (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                           (OperandInt1 "CIA" 0, ""),
                           (OperandString1 "JUMP" (emitLabel (envLabel newEnv)), "OOCV")]
                   ,
                    incEnvLabel newEnv
                   )
                  ]

              _ -> emitError ("operator " ++ operator ++ " not supported")

      (Call (fname, ExpressionEmpty)) ->
              emitExpressionCall0 fname env context
      (Call (fname, (Operation2 (",", expression1, expression2)))) ->
              emitExpressionCall2 fname expression1 expression2 env context
      (Call (fname, expression)) ->
              emitExpressionCall1 fname expression env context

      (ArrayIndex (vname, Number i)) -> 
          case contentTable (envTable env) vname of
            (TypePointer, address) ->
                EmitResult
                (EmitIntValueRegisterA,
                 [(OperandInt1 "TIY" (address + i), "Variable " ++ vname ++ "[" ++ show i ++ "]"),
                  (Operand0 "MA", "")],
                 env)
            _ -> emitError (vname ++ " is not a array")

      (ArrayIndex (vname, expression)) -> 
          let
              emitResult = emitExpression expression env (context {emitContextNeedValue = True})
              newEnv = emitResultEnvironment emitResult
          in
            case contentTable (envTable env) vname of
              (TypePointer, address) ->
                  EmitResult
                  (EmitIntValueRegisterA,
                   emitResultCode emitResult ++
                   (emitValueToRegisterY (emitResultValue emitResult)) ++
                   [(OperandInt1 "AIY" address, "Offset of " ++ vname ++ "[]"),
                    (Operand0 "MA", "")],
                   newEnv)
              _ -> emitError (vname ++ " is not a array")
                  
      _ -> emitError ("unkown expression")

emitExpressionNot (EmitResult (expVal, expCode, expEnv)) context =
    let
        labelNo = envLabel expEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in        
      (EmitResultAlternative
       (EmitResult
        (EmitIntValueRegisterA,
         expCode ++ emitValueToRegisterA expVal ++
         [(OperandInt1 "CIA" 0, "Operator !"),
          (OperandString1 "JUMP" label1, ""),
          (OperandInt1 "TIA" 1, ""),
          (FlagResetRegistersSafe, ""),
          (OperandString1 "JUMP" label2, ""),
          (Label label1, ""),
          (OperandInt1 "TIA" 0, ""),
          (Label label2, "")],
         (incEnvLabel . incEnvLabel) expEnv)
       )
       [
        EmitResult
        (EmitValueFalseLabels [label1],
         expCode ++ emitValueToRegisterA expVal ++
         [(OperandInt1 "CIA" 0, "Operator !"),
          (OperandString1 "JUMP" label1, "OOCV")],
         incEnvLabel expEnv),
        
        EmitResult
        (EmitValueTrueLabels [label1],
         expCode ++ emitValueToRegisterA expVal ++
         [(OperandInt1 "AIA" 0xf, "Operator !"),
          (OperandInt1 "AIA" 1, ""),
          (OperandString1 "JUMP" label1, "OOCV")],
         incEnvLabel expEnv)
       ]
      )

emitExpressionNot (EmitResultAlternative der ers) context =
    (EmitResultAlternative (emitExpressionNot der context) (map reverseEmitValueLabel ers))

reverseEmitValueLabel (EmitResult (EmitValueFalseLabels expLabels, expCode, expEnv)) = (EmitResult (EmitValueTrueLabels expLabels, expCode, expEnv))
reverseEmitValueLabel (EmitResult (EmitValueTrueLabels expLabels, expCode, expEnv)) = (EmitResult (EmitValueFalseLabels expLabels, expCode, expEnv))

emitExpressionAnd expression1 expression2 env context =
    let
        emitResult1 = emitResultSelect emitResultIsFalseLabels (emitExpression expression1 env context)
        emitResult2 = emitResultSelect emitResultIsFalseLabels (emitExpression expression2 (emitResultEnvironment emitResult1) context)
        expVal1 = emitResultValue emitResult1
        expVal2 = emitResultValue emitResult2
        expCode1 = emitResultCode emitResult1
        expCode2 = emitResultCode emitResult2
        newEnv = emitResultEnvironment emitResult2
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
    in
      case (expVal1, expVal2) of
        (EmitValueFalseLabels expLabels1, EmitValueFalseLabels expLabels2) ->
            (EmitResultAlternative
             (EmitResult
              (EmitIntValueRegisterA,
               expCode1 ++ 
               expCode2 ++ 
               [(OperandInt1 "TIA" 1, "Operator &&"),
                (FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "")] ++
               emitLabels (expLabels1 ++ expLabels2) "" ++
               [(OperandInt1 "TIA" 0, ""),
                (Label label1, "")],
               incEnvLabel newEnv)
             )
             [
              EmitResult
              (EmitValueFalseLabels (expLabels1 ++ expLabels2),
               expCode1 ++
               expCode2,
               newEnv)
             ]
            )
            
        _ ->
            let
                emitResult1 = emitExpression expression1 env context
                emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1) context
                expVal1 = emitResultValue emitResult1
                expVal2 = emitResultValue emitResult2
                expCode1 = emitResultCode emitResult1
                expCode2 = emitResultCode emitResult2
                newEnv = emitResultEnvironment emitResult2
                labelNo = envLabel newEnv
                label1 = emitLabel labelNo
                label2 = emitLabel (labelNo + 1)
            in
              (EmitResultAlternative
               (EmitResult
                (EmitIntValueRegisterA,
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandInt1 "TIA" 1, ""),
                  (Label label1, "")],
                 incEnvLabel newEnv)
               )
               [
                EmitResult
                (EmitValueFalseLabels [label1],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "")],
                 incEnvLabel newEnv),
                
                EmitResult
                (EmitValueTrueLabels [label2],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 1"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "AIA" 0xf, "Operator && 2"),
                  (OperandInt1 "AIA" 1, ""),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandString1 "JUMP" label2, ""),
                  (Label label1, "")],
                 (incEnvLabel.incEnvLabel) newEnv)
               ]
              )

emitExpressionOr expression1 expression2 env context =
    let
        emitResult1 = emitResultSelect emitResultIsTrueLabels (emitExpression expression1 env context)
        emitResult2 = emitResultSelect emitResultIsTrueLabels (emitExpression expression2 (emitResultEnvironment emitResult1) context)
        expVal1 = emitResultValue emitResult1
        expVal2 = emitResultValue emitResult2
        expCode1 = emitResultCode emitResult1
        expCode2 = emitResultCode emitResult2
        newEnv = emitResultEnvironment emitResult2
        labelNo = envLabel newEnv
        label1 = emitLabel labelNo
        label2 = emitLabel (labelNo + 1)
    in
      case (expVal1, expVal2) of

        (EmitValueTrueLabels expLabels1, EmitValueTrueLabels expLabels2) ->
            (EmitResultAlternative
             (EmitResult
              (EmitIntValueRegisterA,
               expCode1 ++ 
               expCode2 ++ 
               [(OperandInt1 "TIA" 0, "Operator ||"),
                (FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "")] ++
               emitLabels (expLabels1 ++ expLabels2) "" ++
               [(OperandInt1 "TIA" 1, ""),
                (Label label1, "")],
               incEnvLabel newEnv)
             )
             [
              EmitResult
              (EmitValueTrueLabels (expLabels1 ++ expLabels2),
               expCode1 ++
               expCode2,
               newEnv),

              EmitResult
              (EmitValueFalseLabels [label1],
               expCode1 ++
               expCode2 ++
               [(FlagResetRegistersSafe, ""),
                (OperandString1 "JUMP" label1, "Operator ||")] ++
               emitLabels (expLabels1 ++ expLabels2) "",
               incEnvLabel newEnv)
             ]
            )

        _ ->
            let
                emitResult1 = emitExpression expression1 env context
                emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1) context
                expVal1 = emitResultValue emitResult1
                expVal2 = emitResultValue emitResult2
                expCode1 = emitResultCode emitResult1
                expCode2 = emitResultCode emitResult2
                newEnv = emitResultEnvironment emitResult2
                labelNo = envLabel newEnv
                label1 = emitLabel labelNo
                label2 = emitLabel (labelNo + 1)
            in
              (EmitResultAlternative
               (EmitResult
                (EmitIntValueRegisterA,
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1"),
                  (OperandString1 "JUMP" label1, "")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2"),
                  (OperandString1 "JUMP" label1, ""),
                  (OperandInt1 "TIA" 0, ""),
                  (Label label1, "")],
                 incEnvLabel newEnv)
               )
               [
                EmitResult
                (EmitValueTrueLabels [label1],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1"),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2"),
                  (OperandString1 "JUMP" label1, "")],
                 incEnvLabel newEnv),

                EmitResult
                (EmitValueFalseLabels [label2],
                 expCode1 ++ emitValueToRegisterA expVal1 ++
                 [(OperandInt1 "CIA" 0, "Operator || 1"),
                  (OperandString1 "JUMP" label1, "OOCV")] ++
                 expCode2 ++ emitValueToRegisterA expVal2 ++
                 [(OperandInt1 "CIA" 0, "Operator || 2"),
                  (OperandString1 "JUMP" label1, ""),
                  (FlagResetDontCare, ""),
                  (OperandString1 "JUMP" label2, ""),
                  (Label label1, "")],
                 (incEnvLabel.incEnvLabel) newEnv)
               ]
              )

      
emitDeclarations declarations env = foldl (\x y -> emitDeclaration y x) env declarations

emitDeclaration (Declaration variables) env =
    foldl (\x y -> emitDeclarationVariable y x) env variables

emitDeclarationVariable (Variable (TypeInt, vname)) env =
    if offset - 1 < -1 then
        emitError "memory over"
    else
        Environment (newTable, offset - 1, envLabel env)
    where
      offset = envOffset env
      newTable = updateTable (envTable env) vname (TypeInt, offset)
emitDeclarationVariable (VariableArray (TypeInt, vname, num)) env | num > 0 =
    if offset - num < -1 then
        emitError "memory over"
    else
        Environment (newTable, offset - num, envLabel env)
    where
      offset = envOffset env
      newTable = updateTable (envTable env) vname (TypePointer, offset - num + 1)
emitDeclarationVariable _ env = emitError "invalid declaration"

emitExpressionCall0 fname env context = 
    case fname of
      "exit" ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            EmitResult(EmitVoidValue,
                       [(FlagResetDontCare, ""),
                        (Label label1, ""),
                        (OperandString1 "JUMP" label1, "exit loop")],
                       incEnvLabel env)
      "inputKeyBlock" ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            EmitResult(EmitIntValueRegisterA,
                       [(Label label1, ""),
                        (Operand0 "KA", ""),
                        (OperandString1 "JUMP" label1, "")],
                       incEnvLabel env)
          
      _ ->
          EmitResult (EmitVoidValue, [(FlagResetDontCare, "")] ++ emitFunctionCall0 fname, env)

emitFunctionCall0 fn@"resetNumberLed" = [(OperandString1 "CAL" "RSTO", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepEndSound" = [(OperandString1 "CAL" "ENDS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepErrorSound" = [(OperandString1 "CAL" "ERRS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepShortSound" = [(OperandString1 "CAL" "SHTS", "call " ++ fn ++ "()")]
emitFunctionCall0 fn@"beepLongSound" = [(OperandString1 "CAL" "LONS", "call " ++ fn ++ "()")]
emitFunctionCall0 fname = emitError ("unkown function " ++ fname ++ "(void)")

emitExpressionCall1 fname expression env context = 
    case fname of
      "inputKey" ->
          let
              emitResult = emitExpression expression env (context {emitContextNeedValue = True})
              newEnv = emitResultEnvironment emitResult
              addressCalcCode = emitResultCode emitResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            case addressCalcCode of
              [(OperandInt1 "TIY" _, _)] ->
                  EmitResultAlternative
                  (
                   EmitResult
                   (EmitIntValueRegisterA,
                    [(OperandInt1 "TIA" 0, "inputKey start"),
                     (Operand0 "CH" , ""),
                     (Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", ""),
                     (OperandInt1 "TIA" 1, ""),
                     (Operand0 "CH" , ""),
                     (Label label1, ""),
                     (Operand0 "CH" , "")],
                    incEnvLabel newEnv
                   )
                  )
                  [
                   EmitResult
                   (
                    EmitValueFalseLabels [label1],
                    [(Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "OOCV")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", "")],
                    incEnvLabel newEnv
                   ),
                   EmitResult
                   (
                    EmitValueTrueLabels [label2],
                    [(Operand0 "KA", ""),
                     (OperandString1 "JUMP" label1, "OOCV")] ++
                    addressCalcCode ++
                    [(Operand0 "AM", ""),
                     (FlagResetDontCare, ""),
                     (OperandString1 "JUMP" label2, ""),
                     (Label label1, "")],
                    (incEnvLabel.incEnvLabel) newEnv
                   )
                  ]

              _ -> emitError ("can't emit expression in inputKey(..), code size too large")
      _ ->
          let
              emitResult = emitExpression expression env (context {emitContextNeedValue = True})
          in
            EmitResult(EmitVoidValue,
                       emitResultCode emitResult ++
                       emitValueToRegisterA (emitResultValue emitResult) ++
                       emitFunctionCall1 fname,
                       emitResultEnvironment emitResult)

emitFunctionCall1 fn@"setNumberLed" = [(Operand0 "AO", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"setLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "SETR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"resetLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "RSTR", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"beep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "SUND", "call " ++ fn ++ "()")]
emitFunctionCall1 fn@"sleep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "TIMR", "call " ++ fn ++ "()")]
emitFunctionCall1 fname = emitError ("unkown function " ++ fname ++ "(int)")

emitExpressionCall2 fname expression1 expression2 env context = 
    case fname of
      "setLedBinary" ->
          let
              emitResult1 = emitExpression expression1 env (context {emitContextNeedValue = True})
              emitResult2 = emitExpression expression2 (emitResultEnvironment emitResult1) (context {emitContextNeedValue = True})
              newEnv = emitResultEnvironment emitResult2
          in
            EmitResult(EmitVoidValue,
                       emitResultCode emitResult1 ++
                       (emitValueToRegisterA (emitResultValue emitResult1)) ++
                       [(OperandInt1 "TIY" 0xf, ""), -- ignore "setLedBinary((setLedBinary(i, j),j), i);" pattern
                        (Operand0 "AM", "")] ++
                       emitResultCode emitResult2 ++
                       (emitValueToRegisterA (emitResultValue emitResult2)) ++
                       [(OperandInt1 "TIY" 0xe, ""),
                        (Operand0 "AM", ""),
                        (FlagResetDontCare, ""),         
                        (OperandString1 "CAL" "DSPR", "call " ++ fname ++ "()")],
                       newEnv)
      _ -> emitError ("unkown function " ++ fname ++ "(int, int)")

emitLValue (VariableValue vname) env context =
    case contentTable (envTable env) vname of
      (TypeInt, offset) -> EmitResult (EmitIntVariable offset vname, [], env)
      (TypePointer, address) -> EmitResult (EmitAddress address vname, [], env)

emitLValue (ArrayIndex (vname, Number i)) env context =
    case contentTable (envTable env) vname of
      (TypePointer, address) ->
          EmitResult (EmitAddress (address + i) vname, [], env)
      _ -> emitError (vname ++ " is not a array")

emitLValue (ArrayIndex (vname,expression)) env context =
    let
        emitResult = emitExpression expression env (context {emitContextNeedValue = True})
        newEnv = emitResultEnvironment emitResult
    in
      case contentTable (envTable env) vname of
        (TypePointer, address) ->
            EmitResult
            (EmitAddressRegisterY,
             emitResultCode emitResult ++
             (emitValueToRegisterY (emitResultValue emitResult)) ++
             [(OperandInt1 "AIY" address, "Offset of " ++ vname ++ "[]")],
             newEnv)
        _ -> emitError (vname ++ " is not a array")

emitLValue t _ _ = emitError ("invalid lvalue " ++ show t)

emitValueToRegisterA EmitIntValueRegisterA = []
emitValueToRegisterA (EmitIntValueTemporary i) =
    if i < 0
    then emitError "memory over"
    else [(OperandInt1 "TIY" i, "Temporary offset " ++ show i) , (Operand0 "MA", "")]
emitValueToRegisterA (EmitIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")") , (Operand0 "MA", "")]
emitValueToRegisterA (EmitIntValueConstant i) = 
    [(OperandInt1 "TIA" i, "Constant " ++ show i)]
emitValueToRegisterA EmitIntValueRegisterY = [(Operand0 "CY", "")]
emitValueToRegisterA EmitVoidValue = emitError ("void value not ignored as it ought to be")
emitValueToRegisterA _ = emitError ("invalid value")

emitValueToRegisterY EmitIntValueRegisterY = []
emitValueToRegisterY x = (emitValueToRegisterA x) ++ [(Operand0 "CY", "")]

emitAddressToRegisterY EmitAddressRegisterY = []
emitAddressToRegisterY EmitAddressRegisterA = [(Operand0 "CY", "")]
emitAddressToRegisterY (EmitIntVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")")]
emitAddressToRegisterY (EmitAddress i s) =
    [(OperandInt1 "TIY" i, "Address of '" ++ s ++ "' (offset " ++ show i ++ ")")]
emitAddressToRegisterY t = emitError ("invalid lvalue " ++ show t)

emitLabel = ("LABEL" ++) . show

emitLabels (l:ls) comment = (Label l, comment) : emitLabels (ls) ""
emitLabels [] _ = []

emitResultCodeSizeSort = sortBy emitResultCodeSizeCompare
emitResultCodeSizeCompare (EmitResult (_, a, _)) (EmitResult (_, b, _)) =
    compare (sum (map (instructionSize.fst) a)) (sum (map (instructionSize.fst) b))

immediateValue i | i < 0 = mod (0xf + 1 + i) 0x10
immediateValue i = mod i 0x10


isCalledSetLedBinaryProgram (Program functions) =
    or $ map isCalledSetLedBinaryFunction functions
isCalledSetLedBinaryFunction (Function (fname, ftype, args, block)) = 
    isCalledSetLedBinaryBlock block
isCalledSetLedBinaryBlock (Block (decs, statemants)) = or $ map isCalledSetLedBinaryStatemant statemants

isCalledSetLedBinaryStatemant (If (expression, statement)) =
    isCalledSetLedBinaryExpression expression || isCalledSetLedBinaryStatemant statement
isCalledSetLedBinaryStatemant (IfElse (expression, thenStatement, elseStatement)) =
    isCalledSetLedBinaryExpression expression || isCalledSetLedBinaryStatemant thenStatement ||
    isCalledSetLedBinaryStatemant elseStatement
isCalledSetLedBinaryStatemant (While (expression, statement)) =
    isCalledSetLedBinaryExpression expression || isCalledSetLedBinaryStatemant statement
isCalledSetLedBinaryStatemant (InfiniteLoop statement) =
    isCalledSetLedBinaryStatemant statement
isCalledSetLedBinaryStatemant (StatementBlock block) =
    isCalledSetLedBinaryBlock block
isCalledSetLedBinaryStatemant (StatementExpression expression) =
    isCalledSetLedBinaryExpression expression
isCalledSetLedBinaryStatemant statement = False

isCalledSetLedBinaryExpression (PreOperation1 (op, expression)) = isCalledSetLedBinaryExpression expression
isCalledSetLedBinaryExpression (PostOperation1 (op, expression)) = isCalledSetLedBinaryExpression expression
isCalledSetLedBinaryExpression (Operation2 (op, expression1, expression2)) =
    isCalledSetLedBinaryExpression expression1 || isCalledSetLedBinaryExpression expression2
isCalledSetLedBinaryExpression (Call ("setLedBinary", _)) = True
isCalledSetLedBinaryExpression (Call (fname, expression)) = isCalledSetLedBinaryExpression expression
isCalledSetLedBinaryExpression (ArrayIndex (vname, expression)) = isCalledSetLedBinaryExpression expression
isCalledSetLedBinaryExpression expression = False
